from __future__ import print_function
import gzip
import os
import re
import sys
import io
import zipfile
import tempfile
import shutil

def update_ableton_file(file_path):
    directory, file_name = os.path.split(file_path)
    base_name, ext = os.path.splitext(file_name)
    new_file_path = os.path.join(directory, base_name + '_Synplant2.als')
    if os.path.exists(new_file_path):
        return 0
    with gzip.GzipFile(file_path, 'rb') as f:
        original_data = f.read()
    decompressed_data = original_data.decode('utf-8', errors='strict')
    old_uid_pattern = re.compile(r"<Uid>\s*<Fields\.0\s+Value\s*=\s*\"1448301646\"\s*/>\s*<Fields\.1\s+Value\s*=\s*\"1968394358\"\s*/>\s*<Fields\.2\s+Value\s*=\s*\"1936994911\"\s*/>\s*<Fields\.3\s+Value\s*=\s*\"1718185061\"\s*/>\s*</Uid>")
    new_uid = '<Uid>\n    <Fields.0 Value="1448301646" />\n    <Fields.1 Value="1968394355" />\n    <Fields.2 Value="2037280876" />\n    <Fields.3 Value="1634628608" />\n</Uid>'
    modified_data, replacements_made = old_uid_pattern.subn(new_uid, decompressed_data)
    if replacements_made > 0:
        with gzip.open(new_file_path, 'wt') as f:
            f.write(modified_data.encode('utf-8'))
    return replacements_made

def update_bitwig_file(file_path):
    directory, file_name = os.path.split(file_path)
    base_name, ext = os.path.splitext(file_name)
    new_file_path = os.path.join(directory, base_name + '_Synplant2.bwproject')
    if os.path.exists(new_file_path):
        return 0, 0

    source_sequence_ascii = "5653544E755350767374325F66696C65".encode('ascii')
    target_sequence_ascii = "5653544E75535073796E706C616E7400".encode('ascii')
    with open(file_path, 'rb') as f:
        data = f.read()
    main_file_replacements = data.count(source_sequence_ascii)
    vstpreset_replacements = 0

    if main_file_replacements > 0:
        modified_data = data.replace(source_sequence_ascii, target_sequence_ascii)
        
        if sys.version_info[0] == 2:
            signature_bytes = '504B0304'.decode('hex')
        else:
            signature_bytes = bytes.fromhex('504B0304')

        last_zip_offset = modified_data.rfind(signature_bytes)
        extracted_temp_dir = None

        while last_zip_offset != -1:
            temp_dir = tempfile.mkdtemp()
            try:
                with zipfile.ZipFile(io.BytesIO(modified_data[last_zip_offset:]), 'r') as zip_ref:
                    zip_ref.extractall(temp_dir)
                extracted_temp_dir = temp_dir
                break
            except:
                shutil.rmtree(temp_dir)
                last_zip_offset = modified_data.rfind(signature_bytes, 0, last_zip_offset)

        if extracted_temp_dir is None:
            raise ValueError("No valid ZIP found in data.")

        for root, dirs, files in os.walk(extracted_temp_dir):
            for file in files:
                if file.endswith('.vstpreset'):
                    file_path = os.path.join(root, file)
                    with open(file_path, 'rb') as f:
                        file_data = f.read()
                    if file_data[8:8 + len(source_sequence_ascii)] == source_sequence_ascii:
                        modified_file_data = file_data[:8] + target_sequence_ascii + file_data[8 + len(source_sequence_ascii):]
                        with open(file_path, 'wb') as f:
                            f.write(modified_file_data)
                        vstpreset_replacements += 1

        temp_dir = tempfile.mkdtemp()
        temp_modified_zip_path = os.path.join(temp_dir, 'modified.zip')
        with zipfile.ZipFile(temp_modified_zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for root, dirs, files in os.walk(extracted_temp_dir):
                for file in files:
                    absolute_path = os.path.join(root, file)
                    relative_path = os.path.relpath(absolute_path, extracted_temp_dir)
                    zipf.write(absolute_path, relative_path)

        shutil.rmtree(extracted_temp_dir)

        with open(temp_modified_zip_path, 'rb') as f:
            modified_zip_data = f.read()
        final_data = modified_data[:last_zip_offset] + modified_zip_data
        with open(new_file_path, 'wb') as f:
            f.write(final_data)

        shutil.rmtree(temp_dir)

    return main_file_replacements, vstpreset_replacements

def update_recursively(paths):
    def do_one_file(this_path):
        try:
            if this_path.endswith('.als'):
                replacements_made = update_ableton_file(this_path)
                if replacements_made > 0:
                    replacements_list[this_path] = replacements_made
            elif this_path.endswith('.bwproject'):
                replacements_made, vstpreset_replacements = update_bitwig_file(this_path)
                if replacements_made > 0:
                    replacements_list[this_path] = replacements_made
        except Exception as e:
            errors_list[this_path] = str(e)

    replacements_list = {}
    errors_list = {}
    for this_path in paths:
        if not os.path.isdir(this_path):
            do_one_file(this_path)
        for root, dirs, files in os.walk(this_path):
            for file in files:
                do_one_file(os.path.join(root, file))

    report = ''
    if replacements_list:
        report += "Generated new project files with Synplant version 1 replaced by version 2, for the following projects:\n\n"
        for file_path, replacements_made in replacements_list.items():
            report += "{}: {} {}\n".format(file_path, replacements_made, 'replacement' if replacements_made == 1 else 'replacements')
        report += "-------------------\n"
    if errors_list:
        report += "The following errors occurred:\n\n"
        for file_path, error in errors_list.items():
            report += "{}: {}\n".format(file_path, error)
        report += "-------------------\n"
    report += ('Files generated: ' + str(len(replacements_list)) + '\n')
    report += ('Errors occurred: ' + str(len(errors_list)) + '\n')
    return replacements_list, errors_list, report

if __name__ == "__main__":
    if len(sys.argv) > 1:
        paths = sys.argv[1:]
        replacements_list, errors_list, report = update_recursively(paths)
        print(report)
    else:
        print("Usage: python updateToSynplant2.py <path> [<path> ...]")

